#include <stdio.h>
#include <rpc/rpc.h>
#include <sys/socket.h>
#include <errno.h>

#include "mrpc.h"

extern errno;

/*
 * Ops vector for RMP based rpc service handle
 */
static bool_t		svcrmp_recv();
static enum xprt_stat	svcrmp_stat();
static bool_t		svcrmp_getargs();
static bool_t		svcrmp_reply();
static bool_t		svcrmp_freeargs();
static void		svcrmp_destroy();

static struct xp_ops svcrmp_op = {
  svcrmp_recv,
  svcrmp_stat,
  svcrmp_getargs,
  svcrmp_reply,
  svcrmp_freeargs,
  svcrmp_destroy
};

static int readrmp(), writermp();

struct svc_data {  /* kept in xprt->xp_p1 */
  enum xprt_stat strm_stat;
  u_long x_id;
  XDR xdrs;
  char verf_body[MAX_AUTH_BYTES];
};

SVCXPRT *
svcrmp_create(groupname, sock, sendsize, recvsize)
  char* groupname;
  register int sock;
  u_int sendsize;
  u_int recvsize;
{
  static struct sockaddr_rmp rmpaddr;

  memset((void *)&rmpaddr, 0, sizeof(rmpaddr));
  rmpaddr.srmp_family     = AF_RMP;
  rmpaddr.srmp_min_size   = htonl(1);
  rmpaddr.srmp_min_size   = htonl(1);
  rmpaddr.srmp_ttl        = htonl(1);
  rmpaddr.srmp_port       = htonl(0);
  rmpaddr.srmp_ipport     = htonl(0);
  rmpaddr.srmp_ipaddr     = htonl(0);
  rmpaddr.srmp_unicast    = htonl(RMP_FALSE);
  rmpaddr.srmp_join_flags = htonl(RMP_FORCE_FORM);
  rmpaddr.srmp_qos        = htonl(RMP_QOSTO);
  strcpy(rmpaddr.srmp_groupname, groupname);
  
  return( svcrmp_create_addr(&rmpaddr, sock, sendsize, recvsize) );
}

SVCXPRT *
svcrmp_create_addr(rmpaddr, sock, sendsize, recvsize)
  struct sockaddr_rmp* rmpaddr;
  register int sock;
  u_int sendsize;
  u_int recvsize;
{
  bool_t madesock = RMP_FALSE;
  register SVCXPRT *xprt = (SVCXPRT*)0;
  register struct svc_data *cd;
  struct sockaddr_rmp srmp;
  
  if (sock == RPC_ANYSOCK) {
    if ((sock = socket(AF_RMP, 0, 0)) < 0) {
      fprintf(stderr, "svcrmp_create: rmp socket creation failed\n");
      goto error_handler;
    }
    madesock = RMP_TRUE;
  }
  
  if (connect(sock, (struct sockaddr*)rmpaddr, sizeof(*rmpaddr)) < 0) {
    fprintf(stderr, "svcrmp_create: rmp connect failed\n");
    goto error_handler;
  }
  
  xprt = (SVCXPRT *)mem_alloc(sizeof(SVCXPRT));
  if (xprt == NULL) {
    (void) fprintf(stderr, "svcrmp_create: out of memory\n");
    goto error_handler;
  }
  
  cd = (struct svc_data *)mem_alloc(sizeof(*cd));
  if (cd == NULL) {
    (void) fprintf(stderr, "svcrmp_create: out of memory\n");
    goto error_handler;
  }

  cd->strm_stat = XPRT_IDLE;
  xdrrec_create(&(cd->xdrs), sendsize, recvsize, (caddr_t)xprt, readrmp, writermp);

  xprt->xp_p2 = NULL;
  xprt->xp_p1 = (caddr_t)cd;
  xprt->xp_verf.oa_base = cd->verf_body;
  xprt->xp_addrlen = 0;
  xprt->xp_ops = &svcrmp_op;
  xprt->xp_port = 0;
  xprt->xp_sock = sock;
  xprt_register(xprt);
  
  return( xprt );

 error_handler:
  if (xprt != (SVCXPRT*)0) {
    mem_free((char *) xprt, sizeof(SVCXPRT));
    xprt = (SVCXPRT *)NULL;
  }
  if (madesock) close(sock);
  return( (SVCXPRT*)NULL );
}

static bool_t
svcrmp_recv(xprt, msg)
  SVCXPRT *xprt;
  register struct rpc_msg *msg;
{
  register struct svc_data *cd =
  (struct svc_data *)(xprt->xp_p1);
  register XDR *xdrs = &(cd->xdrs);

  xdrs->x_op = XDR_DECODE;
  (void)xdrrec_skiprecord(xdrs);
  if (xdr_callmsg(xdrs, msg)) {
    cd->x_id = msg->rm_xid;
    return (RMP_TRUE);
  }
  return (RMP_FALSE);
}

static enum xprt_stat
svcrmp_stat(xprt)
  SVCXPRT *xprt;
{
  register struct svc_data *cd =
  (struct svc_data *)(xprt->xp_p1);

  if (cd->strm_stat == XPRT_DIED)
  return (XPRT_DIED);
  if (! xdrrec_eof(&(cd->xdrs)))
  return (XPRT_MOREREQS);
  return (XPRT_IDLE);
}

static bool_t
svcrmp_getargs(xprt, xdr_args, args_ptr)
  SVCXPRT *xprt;
  xdrproc_t xdr_args;
  caddr_t args_ptr;
{
  return ((*xdr_args)(&(((struct svc_data *)(xprt->xp_p1))->xdrs), args_ptr));
}

static bool_t
svcrmp_reply(xprt, msg)
  SVCXPRT *xprt;
  register struct rpc_msg *msg;
{
  register struct svc_data *cd =
  (struct svc_data *)(xprt->xp_p1);
  register XDR *xdrs = &(cd->xdrs);
  register bool_t stat;

  xdrs->x_op = XDR_ENCODE;
  msg->rm_xid = cd->x_id;
  stat = xdr_replymsg(xdrs, msg);
  (void)xdrrec_endofrecord(xdrs, RMP_TRUE);
  return (stat);
}

static void
svcrmp_destroy(xprt)
  register SVCXPRT *xprt;
{
  register struct svc_data *cd = (struct svc_data *)xprt->xp_p1;

  xprt_unregister(xprt);
  (void)close(xprt->xp_sock);
  if (xprt->xp_port != 0) {
    /* a rendezvouser socket */
    xprt->xp_port = 0;
  } else {
    /* an actual connection socket */
    XDR_DESTROY(&(cd->xdrs));
  }
  mem_free((caddr_t)cd, sizeof(struct svc_data));
  mem_free((caddr_t)xprt, sizeof(SVCXPRT));
}

static bool_t
svcrmp_freeargs(xprt, xdr_args, args_ptr)
  SVCXPRT *xprt;
  xdrproc_t xdr_args;
  caddr_t args_ptr;
{
  register XDR *xdrs = &(((struct svc_data *)(xprt->xp_p1))->xdrs);
  
  xdrs->x_op = XDR_FREE;
  return ((*xdr_args)(xdrs, args_ptr));
}

/*
 * All read operations timeout after 100 seconds.
 * A timeout is fatal for the connection.
 */
static struct timeval wait_per_try = { 100, 0 };

/*
 * reads data from the rmp conection.
 * any error is fatal and the connection is closed.
 * (And a read of zero bytes is a half closed stream => error.)
 */
static int
readrmp(xprt, buf, len)
  register SVCXPRT *xprt;
  caddr_t buf;
  register int len;
{
  register int sock = xprt->xp_sock;
#ifdef FD_SETSIZE
  fd_set mask;
  fd_set readfds;

  FD_ZERO(&mask);
  FD_SET(sock, &mask);
#else
  register int mask = 1 << sock;
  int readfds;
#endif				/* def FD_SETSIZE */
  do {
    readfds = mask;
    if (select(FD_SETSIZE, &readfds, (fd_set*)NULL, (fd_set*)NULL, 
	       &wait_per_try) <= 0) {
      if (errno == EINTR) {
	continue;
      }
      goto fatal_err;
    }
  }
#ifdef FD_SETSIZE
  while (!FD_ISSET(sock, &readfds));
#else
  while (readfds != mask);
#endif				/* def FD_SETSIZE */
  if ((len = read(sock, buf, len)) > 0) {
    return (len);
  }
 fatal_err:
  ((struct svc_data *)(xprt->xp_p1))->strm_stat = XPRT_DIED;
  return (-1);
}

/*
 * writes data to the rmp connection.
 * Any error is fatal and the connection is closed.
 */
static int
writermp(xprt, buf, len)
  register SVCXPRT *xprt;
  caddr_t buf;
  int len;
{
  register int i, cnt;
  
  for (cnt = len; cnt > 0; cnt -= i, buf += i) {
    if ((i = write(xprt->xp_sock, buf, cnt)) < 0) {
      ((struct svc_data *)(xprt->xp_p1))->strm_stat =
      XPRT_DIED;
      return (-1);
    }
  }
  return (len);
}
